import tinygradient from "tinygradient";
import {Color as CesiumColor} from "cesium";
import React from "react";
import {AQI_RANGES, getLabelAtValue, range} from "./Helpers";
import {Typography} from "@material-ui/core";
import {HtmlTooltip} from "./HtmlToolip";


type ColorValue = {
    value: number;
    color: string;
}

export class ColorGradient {
    private readonly colorGradient: tinygradient.Instance;
    private readonly values: ColorValue[];
    private readonly min: number;
    private readonly spread: number;

    private customLegend?: JSX.Element;

    get max(): number {return this.min + this.spread;}

    get mid(): number {return this.min + this.spread / 2;}

    constructor(
        public readonly label: string,
        values: ColorValue[],
        private defaultColorCss: string = "#ffffff"
    ) {
        this.values = values.sort((a, b) => a.value - b.value);
        this.spread = values[values.length - 1].value - values[0].value;
        this.min = values[0].value;
        const stops = this.values.map(cv => ({
            color: cv.color, pos: this.normalize(cv.value)
        }));
        this.colorGradient = tinygradient(stops);
    }

    getColorAtAsCssHex(value?: number, defaultColor?: string, darken?: boolean): string {
        if (value === undefined) {
            return defaultColor || this.defaultColorCss;
        }
        let color = this.colorGradient.rgbAt(this.normalize(value));
        if (darken && color.isLight()) {
            color = color.darken();
        }
        return "#" + color.toHex();
    }

    getColorAtAsCesiumColor(value?: number): CesiumColor {
        return CesiumColor.fromCssColorString(this.getColorAtAsCssHex(value));
    }

    getLegendComponent(componentWidth: number = 200) {
        if (this.customLegend) {
            return (<div key={`gradient-${this.label}`} style={{marginBottom: 10}}>
                {this.customLegend}
            </div>);
        }
        const count = 50;
        return (<div key={`gradient-${this.label}`} style={{marginBottom: 10}}>
            <div style={{height: 15, display: "flex", flexDirection: "row"}}>
                <span>{getLabelAtValue(this.min)}</span>
                <span style={{flex: 1}}/>
                <span>{getLabelAtValue(this.mid)}</span>
                <span style={{flex: 1}}/>
                <span>{getLabelAtValue(this.max)}</span>
            </div>
            <div style={{height: 10}}>
                {range(this.min, this.max, count).map(value => (
                    <span key={`grad-${value}`} style={{
                        display: "inline-block",
                        backgroundColor: this.getColorAtAsCssHex(value),
                        width: componentWidth / count,
                        height: 5
                    }}/>
                ))}
            </div>
            <div style={{textAlign: "center", marginTop: 8}}>{this.label}</div>
        </div>);
    }

    setCustomLegend(element: JSX.Element): ColorGradient {
        this.customLegend = element;
        return this;
    }

    private normalize(value: number) {
        return Math.max(Math.min((value - this.min) / this.spread, 1), 0);
    }
}

export const DEFAULT_BUILDINGS_COLOR_CSS = "#d4faff";

export const heatingColorGradient = new ColorGradient(
    "Heating Co2 (kgCO2eq/day)",
    [
        {value: 0, color: "#ffbbbb"},
        {value: 3000, color: "#ff0000"},
    ],
    DEFAULT_BUILDINGS_COLOR_CSS);

export const jamFactorColorGradient = new ColorGradient("Jam Factor", [
    {value: 0, color: "#00ff00"},
    {value: 4, color: "#ffff00"},
    {value: 10, color: "#ff0000"},
]);

export const speedColorGradient = new ColorGradient("Speed (km/h)", [
    {value: 0, color: "#ff0000"},
    {value: 40, color: "#ffff00"},
    {value: 60, color: "#00ff00"},
    {value: 100, color: "#00ff00"},
]);


export const trafficCo2ColorGradient = new ColorGradient("Traffic Co2 (kgCO2eq/day)", [
    {value: 0, color: "#ffbbbb"},
    {value: 3000, color: "#ff0000"}
]);

export const pmTrafficColorGradient = new ColorGradient("Traffic PM2.5 (kg/day)", [
    {value: 0, color: "#00ff00"},
    {value: 150, color: "#ff8c00"},
    {value: 300, color: "#ff0000"},
]);

export const pmSensorColorGradient = new ColorGradient("PM2.5 (µg/m³)", [
    {value: 0, color: "#00ff00"},
    {value: 30, color: "#ff8c00"},
    {value: 100, color: "#ff0000"},
]);

export const aqiSensorColorGradient = new ColorGradient("PM2.5 (µg/m³)",
    AQI_RANGES.flatMap(range => [
        {value: range.aqiRange[0], color: range.color},
        {value: range.aqiRange[1], color: range.color},
    ]),
).setCustomLegend(
    <div>
        <div style={{
            border: "1px solid #adadad",
            borderRadius: 3,
            padding: 5,
            width: 200
        }}>
            {AQI_RANGES.map((aqi, idx) => {
                    let rangeStr = `${aqi.aqiRange[0]} - ${aqi.aqiRange[1]}`;
                    if (idx === AQI_RANGES.length - 1) { // last
                        rangeStr = ` ≥ ${aqi.aqiRange[0]}`;
                    }
                    return <div key={`strategy-${aqi.name}`}
                                style={{display: "flex", alignItems: "center", marginTop: 2}}>
                <span
                    style={{
                        display: "inline-block",
                        width: 12,
                        height: 12,
                        backgroundColor: aqi.color,
                        border: "1px solid black",
                    }}/>
                        <span style={{marginLeft: 5, flex: 1}}>{aqi.name}</span>
                        <HtmlTooltip
                            placement={"top-end"}
                            title={<div>
                                <Typography variant={"body1"}>{aqi.name} ({rangeStr})</Typography>
                                <hr/>
                                <p>{aqi.description}</p>
                            </div>}
                            style={{backgroundColor: "#f5f5f9"}}
                        >
                            <span style={{marginLeft: 10, cursor: "pointer"}}>&#9432;</span>
                        </HtmlTooltip>
                    </div>;
                }
            )}
        </div>
        <div style={{textAlign: "center"}}>Air Quality Index</div>
    </div>
);

// Speed Strategies

type StrategyData = {
    color: string;
    title: string;
    description: JSX.Element;
    range: [number, number];
}

// TODO: fix values
export const speedStrategies: StrategyData[] = [
    {
        color: "#e39d13",
        title: "Low Speeds",
        description: <ul>
            <li>
                <b>Congestion Migration Strategy: </b>
                Increase the average traffic speeds from slower, heavily-congested speeds.
                Examples include include ramp metering and incident management.
            </li>
            <li>
                <b>Traffic Flow Smoothing Techniques: </b>
                Eliminate the stop-and-go effect regarded as shock wave suppression, which also help
                reduce the number of severity of individual accelerations and decelerations.
            </li>
        </ul>,
        range: [0, 55]
    },
    {
        color: "#00ff00",
        title: "Moderate Speeds",
        description: <div>No recommendation.</div>,
        range: [55, 105]
    },
    {
        color: "#ff0000",
        title: "Very High Speeds",
        description: <div>
            <b>Speed Management Techniques: </b>
            Reduce excessively high speeds to safe speeds. E.g. direct enforcement by police,
            radar camera, or aircraft.
        </div>,
        range: [105, 150]
    },
];

export const speedStrategyColorGradient = new ColorGradient(
    "Speed Strategy",
    speedStrategies.flatMap(strategy => [
        {value: strategy.range[0] + 0.001, color: strategy.color},
        {value: strategy.range[1], color: strategy.color}
    ])
).setCustomLegend(
    <div>
        {speedStrategies.map((strategy, idx) => {
                let speedRange = `${strategy.range[0]} - ${strategy.range[1]} km/h`;
                if (idx === speedStrategies.length - 1) { // last
                    speedRange = ` > ${strategy.range[0]} km/h`;
                }
                return <div key={`strategy-${strategy.title}`}
                            style={{display: "flex", alignItems: "center", marginTop: 2}}>
                <span
                    style={{
                        display: "inline-block",
                        width: 12,
                        height: 12,
                        backgroundColor: strategy.color,
                        border: "1px solid black",
                    }}/>
                    <span style={{marginLeft: 5}}>{strategy.title}</span>
                    <span style={{flex: 1}}/>
                    <HtmlTooltip
                        placement={"top-end"}
                        title={<div>
                            <Typography variant={"body1"}>{strategy.title} ({speedRange})</Typography>
                            <hr/>
                            <p>{strategy.description}</p>
                        </div>}
                        style={{backgroundColor: "#f5f5f9"}}
                    >
                        <span style={{marginLeft: 10, cursor: "pointer"}}>&#9432;</span>
                    </HtmlTooltip>
                </div>;
            }
        )}

    </div>
);

