import {getObservationUrlForFeatureOfInterest, getTrafficGeoJsonUrlForFeaturesWithinBBox, getTrafficHistoryUrl} from "./TrafficConfig";
import {FeatureCollection, GeoJsonProperties, Point} from "geojson";
import {getPMGeoJsonUrlWithLastObservation, getPMHistoryUrl} from "./PMConfig";
import {parseISO, subHours} from "date-fns";
import {getAQIFromPM} from "./Helpers";

const PM_RESULT_POST_KEY = "result";
const TRAFFIC_LENGTH_KEY = "properties/length";

export async function getTrafficGeoJsonWithAverages(
    fromDate?: Date,
    toDate?: Date,
    fromSpeed?: number,
    toSpeed?: number,
    fromJamFactor?: number,
    toJamFactor?: number,
): Promise<{ geoJson: FeatureCollection, isFiltered: boolean }> {
    const filtered = ![fromDate, toDate, fromSpeed, toSpeed, fromJamFactor, toJamFactor].every(x => x === undefined);
    const limitObservations = filtered ? 1000 : 1;

    const geoJsonUrl = getTrafficGeoJsonUrlForFeaturesWithinBBox();
    return fetch(geoJsonUrl)
        .then(res => res.json())
        .then((geoJson: FeatureCollection) => {
            const promises = geoJson.features.map(async feature => {
                if (!feature.properties)
                    return;
                const id = feature.properties.id;
                const additionalProperties = await getAveragesForFeatureOfInterest(
                    id,
                    limitObservations,
                    fromDate,
                    toDate,
                    fromSpeed,
                    toSpeed,
                    fromJamFactor,
                    toJamFactor,
                );

                feature.properties = {
                    id: feature.properties.id,
                    name: feature.properties.name,
                    length: feature.properties[TRAFFIC_LENGTH_KEY],
                    description: feature.properties.description,
                    ...additionalProperties
                };
            });

            return Promise.all(promises).then(() => {
                geoJson.features = geoJson.features.filter(feature => feature.properties?.hasData);
                return {geoJson, isFiltered: filtered};
            });
        });
}

export async function getAveragesForFeatureOfInterest(
    id: string,
    limitObservations: number,
    fromDate?: Date,
    toDate?: Date,
    fromSpeed?: number,
    toSpeed?: number,
    fromJamFactor?: number,
    toJamFactor?: number,
) {
    type ItemType = { resultQuality: number, result: number, parameters: { jamFactor: number } }
    let hasNext = true;
    let data: ItemType[] = [];
    let skipObservations = 0;
    let maxPages = 3;
    while (hasNext) {
        const response = await fetch(getObservationUrlForFeatureOfInterest(
            id,
            fromDate,
            toDate,
            fromSpeed,
            toSpeed,
            fromJamFactor,
            toJamFactor,
            limitObservations,
            skipObservations,
        )).then(res => res.json());
        data = data.concat(response.value);
        maxPages--;
        if (limitObservations < 1000 || !response["@iot.nextLink"] || !maxPages) {
            hasNext = false;
        } else {
            skipObservations += limitObservations;
        }
    }
    console.log(data);
    return Promise.resolve(data)
        .then((values: ItemType[]) => {
            const speedAcc = new Accumulator();
            const qualityAcc = new Accumulator();
            const jamAcc = new Accumulator();
            if (!values.length) {
                return {
                    hasData: false
                };
            }
            values.forEach(({result, parameters, resultQuality}) => {
                speedAcc.accumulate(result);
                qualityAcc.accumulate(resultQuality);
                jamAcc.accumulate(parameters.jamFactor);
            });
            return {
                hasData: true,
                speed: speedAcc.getAverage(),
                jamFactor: jamAcc.getAverage(),
                quality: qualityAcc.getAverage(),
            };
        });
}

type TrafficistoryResult = {
    phenomenonTime: Date;
    speed: number;
    jam: number;
};

export async function getTrafficHistory(
    id: string,
    fromDate: Date,
    toDate: Date,
): Promise<TrafficistoryResult[]> {
    const url = getTrafficHistoryUrl(
        id,
        fromDate,
        toDate,
        5000,
    );

    return fetch(url)
        .then(res => res.json())
        .then((json: any) => json.value)
        .then((values: { phenomenonTime: string, result: number, parameters: { jamFactor: number } }[]) =>
            values.map(it => ({
                phenomenonTime: parseISO(it.phenomenonTime),
                speed: it.result,
                jam: it.parameters.jamFactor,
            }))
        );
}

export async function getPMGeoJsonWithAverages(
    fromDate?: Date,
    toDate?: Date,
    fromPMValue?: number,
    toPMValue?: number,
): Promise<{ geoJson: FeatureCollection, isFiltered: boolean }> {
    const filtered = ![fromDate, toDate, fromPMValue, toPMValue].every(x => x === undefined);
    const limitObservations = filtered ? 100 : 1;

    if (!fromDate && !toDate) {
        toDate = new Date();
        fromDate = subHours(toDate, 1);
    }

    const url = getPMGeoJsonUrlWithLastObservation(
        fromDate,
        toDate,
        fromPMValue,
        toPMValue,
        limitObservations
    );

    return fetch(url)
        .then(res => res.json())
        .then((geoJson: FeatureCollection) => {
            geoJson.features.forEach(feature => {
                const geometry = feature.geometry as Point;
                const [latitude, longitude] = geometry.coordinates;
                geometry.coordinates = [longitude, latitude];

                const pm = getPMAverage(feature.properties);
                const aqi = getAQIFromPM(pm);
                if (!feature.properties)
                    return;
                feature.properties = {
                    id: feature.properties.id,
                    name: feature.properties.name,
                    description: feature.properties.description,
                    longitude,
                    latitude,
                    pm,
                    aqi
                };
            });
            return {geoJson, isFiltered: filtered};
        });
}


type PMHistoryResult = { phenomenonTime: Date, result: number };

export async function getPMHistory(
    id: number,
    fromDate: Date,
    toDate: Date,
): Promise<PMHistoryResult[]> {
    const url = getPMHistoryUrl(
        id,
        fromDate,
        toDate,
        1000
    );

    return fetch(url)
        .then(res => res.json())
        .then((json: any) => json.value)
        .then((values: { phenomenonTime: string, result: number }[]) =>
            values.map(it => ({
                phenomenonTime: parseISO(it.phenomenonTime),
                result: it.result
            }))
        );
}

class Accumulator {
    private count: number = 0;
    private sum: number = 0;

    accumulate(value: number) {
        this.count++;
        this.sum += value;
    }

    getAverage(): number {
        return this.sum / this.count;
    }
}

function getPMAverage(properties: GeoJsonProperties) {
    if (!properties)
        throw new Error("Unexpected null properties");
    const pmAcc = new Accumulator();
    Object.entries(properties).forEach(([key, property]) => {
            if (key.endsWith(PM_RESULT_POST_KEY)) {
                pmAcc.accumulate(+property);
            }
        }
    );
    return pmAcc.getAverage();
}
