import {Cartesian3} from "cesium";
import {format} from "date-fns";

const AVERAGE_CARS_PER_DAY_IN_STUTTGART = 62_000;

export const BLACK_CSS = "black";

export type Dictionary<T> = {
    [key: string]: T
}

export function deepClone<T>(obj: T): T {
    return JSON.parse(JSON.stringify(obj));
}

export function arraySum(arr: number[]): number {
    return arr.reduce(function (a, b) {
        return a + b;
    }, 0);
}

export function arrayAverage(arr: number[]): number {
    return arr.length ? arraySum(arr) / arr.length : 0;
}

export function arrayMin(arr: number[]): number | undefined {
    return arr.length ? arr.reduce(function (a, b) {
        return a < b ? a : b;
    }) : undefined;
}

export function arrayMax(arr: number[]): number | undefined {
    return arr.length ? arr.reduce(function (a, b) {
        return a > b ? a : b;
    }) : undefined;
}

export function randomString(length: number) {
    let result = "";
    const characters = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
    const charactersLength = characters.length;
    for (let i = 0; i < length; i++) {
        result += characters.charAt(Math.floor(Math.random() * charactersLength));
    }
    return result;
}

export function sleep(ms: number): Promise<void> {
    return new Promise(resolve => setTimeout(resolve, ms));
}

export function chunkArray<T>(arr: T[], size: number): T[][] {
    return Array.from({length: Math.ceil(arr.length / size)}, (v, i) =>
        arr.slice(i * size, i * size + size)
    );
}

export function distanceSquaredInMeters(lon1: number, lat1: number, lon2: number, lat2: number): number {
    const start = Cartesian3.fromDegrees(lon1, lat1);
    const end = Cartesian3.fromDegrees(lon2, lat2);
    return Cartesian3.distanceSquared(start, end);
}

export function encodeFrostId(id: string) {
    return id.replaceAll("+", "%2B");
}

export function roundTo(n?: number, decimalPoints: number = 2): number | undefined {
    if (n === undefined) return undefined;
    const factor = 10 ** decimalPoints;
    return Math.round(n * factor) / factor;
}

export function formatTimestamp(timestamp: number, formatStr = "dd/MM HH:mm") {
    const d = new Date(timestamp);
    try {
        return format(d, formatStr);
    } catch {
        return "";
    }
}

export function numberWithCommas(n?: number, roundToDecimalPoints: number = 2): string | undefined {
    n = roundTo(n, roundToDecimalPoints);
    return n?.toString().replace(/\B(?=(\d{3})+(?!\d))/g, ",");
}

export function kgCo2PerDayFactorFromSpeed(speed: number, segmentLength: number) {
    return speed
        * segmentLength
        * 0.5302 // Factor
        * 0.001 // g to kg
        * AVERAGE_CARS_PER_DAY_IN_STUTTGART;
}

export function pmFactorFromSpeed(speed: number, segmentLength: number) {
    return speed
        * segmentLength
        * 0.04 // Factor
        * 0.001 // g to kg
        * AVERAGE_CARS_PER_DAY_IN_STUTTGART;
}

export function getLabelAtValue(value?: number): string | undefined {
    if (value === undefined)
        return undefined;
    const ONE_MILLION = 1_000_000;
    const ONE_THOUSAND = 1_000;
    if (value >= ONE_MILLION) {
        return Math.floor(10 * value / ONE_MILLION) / 10 + "M";
    }
    if (value >= ONE_THOUSAND) {
        return Math.floor(10 * value / ONE_THOUSAND) / 10 + "k";
    }
    return value.toString();
}

export function range(from: number, to: number, count: number = 10) {
    const step = (to - from) / (count - 1);
    return [...Array(count)].map((_, i) => from + i * step);
}

type Range = [number, number];
export type AQIRange = {
    name: string;
    description: string;
    color: string;
    aqiRange: Range;
    pmRange: Range;
}
export const AQI_RANGES: AQIRange[] = [
    {
        aqiRange: [0, 50],
        pmRange: [0, 10],
        color: "#00ff00",
        name: "Good",
        description: "Air quality is considered satisfactory, and air pollution poses little or no risk.",
    },
    {
        aqiRange: [51, 100],
        pmRange: [11, 20],
        color: "#ffff00",
        name: "Moderate",
        description: "Air quality is acceptable; however, for some pollutants there may be a moderate health concern for a very small number " +
            "of people who are unusually sensitive to air pollution.",
    },
    {
        aqiRange: [101, 150],
        pmRange: [21, 25],
        color: "#ff8400",
        name: "Unhealthy for Sensitive Groups",
        description: "Members of sensitive groups may experience health effects. The general public is not likely to be affected.",
    },
    {
        aqiRange: [151, 200],
        pmRange: [26, 50],
        color: "#ff0000",
        name: "Unhealthy",
        description: "Everyone may begin to experience health effects; members of sensitive groups may experience more serious health effects.",
    },
    {
        aqiRange: [201, 300],
        pmRange: [51, 100],
        color: "#b103b1",
        name: "Very Unhealthy",
        description: "Health warnings of emergency conditions. The entire population is more likely to be affected.",
    },
];

function isInRange(n: number, range: Range) {
    return n >= range[0] && n <= range[1];
}

/**
 * @param pm in µg/m³
 */
export function getAQIFromPM(pm: number): { aqiValue: number, aqiRange: AQIRange } {
    pm = Math.ceil(pm);
    let range = AQI_RANGES.find(range => isInRange(pm, range.pmRange));
    if (!range)
        range = AQI_RANGES.reduce((range, acc) =>
            range.aqiRange[0] >= acc.aqiRange[0] ? range : acc
        );
    const [aqiLo, aqiHi] = range.aqiRange;
    const [pmLo, pmHi] = range.pmRange;
    const slope = (aqiHi - aqiLo) / (pmHi - pmLo);
    return {aqiValue: Math.ceil(slope * (pm - pmLo) + aqiLo), aqiRange: range};
}
