import {arrayAverage, kgCo2PerDayFactorFromSpeed, Dictionary, pmFactorFromSpeed} from "./Helpers";
import {Cesium3DTile, Cesium3DTileFeature, Ellipsoid, Entity as CesiumEntity, Math as CesiumMath} from "cesium";


type LonLat = {
    longitude: number;
    latitude: number;
}

export type TrafficCacheInfo = {
    description: string;
    points: LonLat[]
    midPoint: LonLat;
    /**
     * The point 200 meters below and 200 meters to the left of the midpoint
     */
    bottomLeft200: LonLat;
    /**
     * The point 200 meters above and 200 meters to the right of the midpoint
     */
    topRight200: LonLat;
    speed: number;
    pmFactor: number;
    jamFactor: number;
    co2Factor: number;
    quality: number;
    segmentLength: number; // In Kilometers
};

export type TileFeaturesCacheInfo = {
    parentId: string;
    longitude: number;
    latitude: number;
    isCloseToStreet?: boolean;
    dirty: boolean;
};

const FEATURE_ID_PROPERTY_NAME = "gml_id";
const FEATURE_PARENT_ID_PROPERTY_NAME = "gml_parent_id";

const TRAFFIC_DESCRIPTION_KEY = "description";
const TRAFFIC_SPEED_KEY = "speed";
const TRAFFIC_JAM_FACTOR_KEY = "jamFactor";
const TRAFFIC_QUALITY_KEY = "quality";
const TRAFFIC_LENGTH_KEY = "length";


function getFeatureId(feature: Cesium3DTileFeature): string | undefined {
    return feature.getProperty && feature.getProperty(FEATURE_ID_PROPERTY_NAME);
}

function getFeatureParentId(feature: Cesium3DTileFeature): string | undefined {
    return feature.getProperty && feature.getProperty(FEATURE_PARENT_ID_PROPERTY_NAME);
}

export class TrafficCacheManager {
    trafficCache: Dictionary<TrafficCacheInfo> = {};
    tileFeaturesCache: Dictionary<TileFeaturesCacheInfo> = {};

    getIsCloseToStreet(feature: Cesium3DTileFeature): boolean | undefined {
        const tileInfo = this.getTilesInfo(feature);
        return tileInfo?.dirty ? undefined : tileInfo?.isCloseToStreet;
    }

    hasTrafficInfo(): boolean {
        return !!Object.keys(this.trafficCache).length;
    }

    recalculateTrafficCache(cesiumEntities?: CesiumEntity[]) {
        this.trafficCache = {};
        cesiumEntities?.forEach(entity => {
            if (entity.properties) {
                const twoHundredMetersInDegrees = 200 / 111000;
                // @ts-ignore
                const locations = Ellipsoid.WGS84.cartesianArrayToCartographicArray(entity.polyline.positions._value),
                    points = locations.map(p => ({longitude: CesiumMath.toDegrees(p.longitude), latitude: CesiumMath.toDegrees(p.latitude)})),
                    longitude = arrayAverage(points.map(p => p.longitude)),
                    latitude = arrayAverage(points.map(p => p.latitude)),
                    bottomLeft200: LonLat = {longitude: longitude - twoHundredMetersInDegrees, latitude: latitude - twoHundredMetersInDegrees},
                    topRight200: LonLat = {longitude: longitude + twoHundredMetersInDegrees, latitude: latitude + twoHundredMetersInDegrees},
                    description = entity.properties[TRAFFIC_DESCRIPTION_KEY].toString(),
                    speed = +entity.properties[TRAFFIC_SPEED_KEY],
                    jamFactor = +entity.properties[TRAFFIC_JAM_FACTOR_KEY],
                    quality = +entity.properties[TRAFFIC_QUALITY_KEY],
                    segmentLength = +entity.properties[TRAFFIC_LENGTH_KEY],
                    co2Factor = kgCo2PerDayFactorFromSpeed(speed, segmentLength),
                    pmFactor = pmFactorFromSpeed(speed, segmentLength);

                this.trafficCache[entity.name!] = {
                    description,
                    points,
                    midPoint: {longitude, latitude},
                    bottomLeft200,
                    topRight200,
                    speed,
                    jamFactor,
                    co2Factor,
                    quality,
                    segmentLength,
                    pmFactor,
                };
            }
        });
    }

    setTileInfo(tile: Cesium3DTile) {
        for (let i = 0; i < tile.content.featuresLength; i++) {
            const feature = tile.content.getFeature(i);
            this.setInitialFeatureInfo(feature);
        }
    }

    setInitialFeatureInfo(feature: Cesium3DTileFeature) {
        const longitude = feature.getProperty("longtitude") || feature.getProperty("longitude");
        const latitude = feature.getProperty("latitude");

        if (!longitude || !latitude)
            return;

        this.tileFeaturesCache[getFeatureId(feature)!] = {
            parentId: feature.getProperty(FEATURE_PARENT_ID_PROPERTY_NAME),
            longitude: +longitude,
            latitude: +latitude,
            dirty: true
        };
    }

    async makeAllFeaturesDirty() {
        Object.values(this.tileFeaturesCache)
            .forEach((feature) => {
                feature.dirty = true;
            });
    }

    async recalculateFactorsForDirtyFeatures() {
        return Promise.all(
            Object.values(this.tileFeaturesCache)
                .filter((feature) => feature.dirty)
                .map(async (feature) => {
                    const parent = this.getTilesInfoByd(feature.parentId);
                    if (parent && !parent.dirty) {
                        feature.isCloseToStreet = parent.isCloseToStreet;
                    } else {
                        feature.isCloseToStreet = this.isCloserThan200MetersToAStreet(feature.longitude, feature.latitude);
                    }
                    feature.dirty = false;
                })
        ).then(() => {});
    }

    private isCloserThan200MetersToAStreet(featureLongitude: number, featureLatitude: number) {
        // const distanceToFeature = (lon: number, lat: number) =>
        //     distanceSquaredInMeters(lon, lat, featureLongitude, featureLatitude);

        for (const t of Object.values(this.trafficCache)) {
            if (featureLongitude <= t.topRight200.longitude &&
                featureLatitude <= t.topRight200.latitude &&
                featureLongitude >= t.bottomLeft200.longitude &&
                featureLatitude >= t.bottomLeft200.latitude)
                return true;
            // const d = distanceToFeature(t.midPoint.longitude, t.midPoint.latitude);
            // if (d <= 40000) {
            //     return true;
            // }
        }
        return false;
    }

    getTilesInfo(feature: Cesium3DTileFeature): TileFeaturesCacheInfo | undefined {
        if (!feature.getProperty) // not a feature
            return undefined;
        const id = getFeatureId(feature);
        const parentId = getFeatureParentId(feature);
        const tilesInfo = this.getTilesInfoByd(parentId) || this.getTilesInfoByd(id);
        if (tilesInfo)
            return tilesInfo;
        this.setInitialFeatureInfo(feature);
        return this.getTilesInfoByd(id);

    }

    private getTilesInfoByd(featureId?: string): TileFeaturesCacheInfo | undefined {
        return featureId ? this.tileFeaturesCache[featureId] : undefined;
    }
}
