import React, {ChangeEvent} from "react";
import {
    AccordionDetails,
    Button,
    Checkbox,
    FormControlLabel,
    IconButton,
    TextField,
    Tooltip,
    Typography
} from "@material-ui/core";
import {CenterFocusWeak, DoubleArrow, ExpandMore, OpenInNew} from "@material-ui/icons";
import {_3DContainer, Collections, Type} from "ogc3dcontainerentitiesmodel";
import {TilesetInfo} from "../Map";
import {TreeView} from "@material-ui/lab";
import {MinusSquare, PlusSquare, StyledTreeItem} from "../StyledTreeItems";
import {StyledAccordion, StyledAccordionSummary} from "../StyledAccordion";

type ContainerProps = {
    onTilesetAdded: (tilesetInfo: TilesetInfo) => void;
    onTilesetRemoved: (id: string) => void;
    getVisibleItemsIds: () => string[];
    flyToTileset: (tilesetId: string) => void;
}

type ContainerState = {
    url: string,
    collections: Collections | null
}

export class OGC3DContainerLayer extends React.Component<ContainerProps, ContainerState> {
    constructor(props: ContainerProps) {
        super(props);
        this.state = {
            url: "http://localhost:3001/collections",
            collections: null
        };
    }

    handleUrlChange(element: ChangeEvent<HTMLInputElement | HTMLTextAreaElement>) {
        this.setState({url: element.target.value});
    }

    loadData() {
        fetch(this.state.url)
            .then(response => response.json())
            .then(data => this.setState({collections: Collections.fromJson(data)}))
            .catch(err => {
                this.setState({collections: null});
                console.error(err);
            });
    }

    render() {
        const {url, collections} = this.state;
        return <StyledAccordion square>
            <StyledAccordionSummary expandIcon={<ExpandMore/>}>
                <Typography>3D Container</Typography>
            </StyledAccordionSummary>
            <AccordionDetails style={{flexDirection: "column"}}>
                <div style={{padding: 5, display: "flex", flex: 1}}>
                    <TextField style={{flex: 1}} label={"3D Container Url"} value={url} onChange={this.handleUrlChange.bind(this)}/>
                    <IconButton type="submit" aria-label="search" onClick={this.loadData.bind(this)}>
                        <DoubleArrow/>
                    </IconButton>
                </div>
                {collections?.collections ?
                    <TreeView
                        defaultCollapseIcon={<MinusSquare/>}
                        defaultExpandIcon={<PlusSquare/>}>
                        {collections.collections.map(collection => (
                            <CollectionInfo
                                deepId={collection.id}
                                key={collection.id}
                                collection={collection}
                                onTilesetAdded={this.props.onTilesetAdded}
                                onTilesetRemoved={this.props.onTilesetRemoved}
                                getVisibleItemsIds={this.props.getVisibleItemsIds}
                                flyToTileset={this.props.flyToTileset}
                            />
                        ))}
                    </TreeView>

                    : null
                }
            </AccordionDetails>
        </StyledAccordion>;
    }
}

type CollectionProps = {
    deepId: string;
    collection: _3DContainer;
    onTilesetAdded: (tilesetInfo: TilesetInfo) => void;
    onTilesetRemoved: (id: string) => void;
    getVisibleItemsIds: () => string[];
    flyToTileset: (tilesetId: string) => void;
}

type CollectionState = {}

class CollectionInfo extends React.Component<CollectionProps, CollectionState> {

    render() {
        const {collection} = this.props;
        const deepId = `${this.props.deepId}`;
        return <StyledTreeItem nodeId={deepId} label={collection.title}>
            {collection?.children?.length ? (
                collection.children.map(subCollection => (
                    <CollectionInfo
                        key={`${deepId}-${subCollection.id}`}
                        deepId={`${deepId}-${subCollection.id}`}
                        collection={subCollection}
                        onTilesetAdded={this.props.onTilesetAdded}
                        onTilesetRemoved={this.props.onTilesetRemoved}
                        getVisibleItemsIds={this.props.getVisibleItemsIds}
                        flyToTileset={this.props.flyToTileset}
                    />
                ))
            ) : (<div
                style={{
                    display: "flex",
                    justifyContent: "space-evenly",
                    alignItems: "center"
                }}>
                <CollectionContent
                    deepId={`${deepId}-${collection.id}`}
                    key={`${deepId}-${collection.id}`}
                    collection={collection}
                    onTilesetAdded={this.props.onTilesetAdded}
                    onTilesetRemoved={this.props.onTilesetRemoved}
                    getVisibleItemsIds={this.props.getVisibleItemsIds}
                    flyToTileset={this.props.flyToTileset}
                />
            </div>)
            }
        </StyledTreeItem>;
    }
}

type CheckedDictionary = { [tilesetId: string]: boolean }

type CollectionContentState = {
    checked: CheckedDictionary;
}

type CollectionContentProps = {
    deepId: string;
    collection: _3DContainer;
    onTilesetAdded: (tilesetInfo: TilesetInfo) => void;
    onTilesetRemoved: (id: string) => void;
    getVisibleItemsIds: () => string[];
    flyToTileset: (tilesetId: string) => void;
}

class CollectionContent extends React.Component<CollectionContentProps, CollectionContentState> {
    constructor(props: CollectionContentProps) {
        super(props);
        this.state = {
            checked: this.resolveChecked()
        };
    }

    private resolveChecked(): CheckedDictionary {
        return this.props.collection.content?.reduce((obj, link) => {
            const id = this.getTilesetId(link.type);
            return {...obj, [id]: this.isChecked(id)};
        }, {}) || {};
    }

    handleCheckboxChange(tileset: TilesetInfo, checked: boolean) {
        if (checked)
            this.props.onTilesetAdded(tileset);
        else
            this.props.onTilesetRemoved(tileset.id);
        this.setState(state => ({
            checked: {
                ...state.checked,
                [tileset.id]: checked
            }
        }));
    }

    isChecked(tilesetId: string) {
        return this.props.getVisibleItemsIds().some(id => id === tilesetId);
    }

    getTilesetId(linkType?: Type) {
        return `${this.props.deepId}-${linkType}`;
    }

    render() {
        const {collection} = this.props;
        return collection.content?.map(link => {
            const controlProvider = (linkType?: Type) => {
                const tilesetId = this.getTilesetId(linkType);
                // noinspection JSUnreachableSwitchBranches
                switch (linkType) {
                    // @ts-ignore
                    case "application/json+3dtiles":
                    case "application/json-3dtiles":
                        return (<span>
                            <FormControlLabel
                                label="3D Tiles"
                                control={
                                    <Checkbox
                                        color={"primary"}
                                        checked={this.state.checked[tilesetId]}
                                        onChange={(e) => this.handleCheckboxChange(
                                            {
                                                id: tilesetId,
                                                link
                                            },
                                            e.target.checked)}/>
                                }/>
                            <Tooltip title={"Fly to"} style={{visibility: this.state.checked[tilesetId] ? "unset" : "hidden"}}>
                                <span>
                                    <IconButton
                                        size={"small"}
                                        onClick={() => this.props.flyToTileset(tilesetId)}
                                    >
                                    <CenterFocusWeak/>
                                </IconButton>
                                </span>
                            </Tooltip>
                        </span>);
                    default:
                        return (<Button
                            size={"small"}
                            endIcon={<OpenInNew/>}
                            variant={"contained"}
                            color={"primary"}
                            target="_blank"
                            rel="noopener noreferrer"
                            href={link.href!}>{linkType === "application/json-i3s" ? "I3S" : linkType}</Button>);
                }

            };
            return <span key={`${collection.id}-${link.type}`}>
                {controlProvider(link.type)}
            </span>;
        }) || (<div>No Content!</div>);
    }
}
