import React from "react";
import {Button, Checkbox, Divider, FormControlLabel, FormGroup, Grid, Switch, TextField, Typography} from "@material-ui/core";
import {DateTimePicker, DatePicker, MuiPickersUtilsProvider} from "@material-ui/pickers";
import DateFnsUtils from "@date-io/date-fns";
import {DateFilter, TrafficVisibilityMode} from "../Map";
import {getPMGeoJsonWithAverages, getTrafficGeoJsonWithAverages} from "../../common/FrostApiProxy";
import {FeatureCollection} from "geojson";
import {Color as AlertSeverity} from "@material-ui/lab/Alert/Alert";
import {endOfDay, startOfDay, subWeeks} from "date-fns";

type TrafficLayerProps = {
    onTrafficLayerAdded: (geoJson: FeatureCollection, mode: TrafficVisibilityMode, dateFilter: DateFilter, isFiltered: boolean) => void;
    onPMLayerAdded: (geoJson: FeatureCollection, mode: TrafficVisibilityMode, dateFilter: DateFilter, isFiltered: boolean) => void;
    onTrafficLayerRemoved: () => void;
    onColorByStrategyChanged: (value: boolean) => void;
    onColorByAqiChanged: (value: boolean) => void;
    addLoadingId: (id: string) => void;
    removeLoadingId: (id: string) => void;
    showSnackbar: (message: string, severity: AlertSeverity) => void;
}

type TrafficLayerState = {
    minSpeed?: number;
    maxSpeed?: number;
    minJamFactor?: number;
    maxJamFactor?: number;
    minPMValue?: number;
    maxPMValue?: number;
    minDate: Date | null;
    maxDate: Date | null;
    day: Date | null;
    visibleLayer?: TrafficVisibilityMode,
    isLoadingTraffic: boolean;
    isLoadingPM: boolean;
    colorSpeedByStrategy: boolean;
    colorPMByAqi: boolean;
}

function stringToNumberOrUndefined(value?: string): number | undefined {
    return value !== undefined && value.length ? +value : undefined;
}

function numberOrEmptyString(value?: number): number | string {
    return value !== undefined ? value : "";
}

export class STALayer extends React.Component<TrafficLayerProps, TrafficLayerState> {
    constructor(props: TrafficLayerProps) {
        super(props);
        this.state = {
            minDate: null,
            maxDate: null,
            day: null,
            isLoadingTraffic: false,
            isLoadingPM: false,
            colorSpeedByStrategy: false,
            colorPMByAqi: false,
        };
    }

    get isLoading() {
        return this.state.isLoadingPM || this.state.isLoadingTraffic;
    }

    private getSTADates(mode: TrafficVisibilityMode) {
        if (mode === "co2" || mode === "pm") {
            const maxDate = endOfDay(this.state.day || new Date());
            const minDate = startOfDay(this.state.day || new Date());
            return {
                minDate: minDate,
                maxDate: maxDate,
            };
        }
        return {
            minDate: this.state.minDate || undefined,
            maxDate: this.state.maxDate || undefined,
        };
    }

    private async handleSubmitClick(mode?: TrafficVisibilityMode): Promise<void> {
        mode = mode || this.state.visibleLayer;
        if (!mode)
            return;
        if (mode !== this.state.visibleLayer) {
            this.props.onTrafficLayerRemoved();
        }

        const {minDate, maxDate} = this.getSTADates(mode);
        if (!!minDate !== !!maxDate) {
            this.props.showSnackbar("Either both dates or no date should be chosen!", "info");
            return;
        }

        const trafficLayerFetchId = "traffic-layer-fetch";
        this.props.addLoadingId(trafficLayerFetchId);
        this.setState({isLoadingTraffic: true});
        const promises: Promise<void>[] = [];

        promises.push(
            getTrafficGeoJsonWithAverages(
                minDate,
                maxDate,
                mode === "speed" ? this.state.minSpeed : undefined,
                mode === "speed" ? this.state.maxSpeed : undefined,
                mode === "jam" ? this.state.minJamFactor : undefined,
                mode === "jam" ? this.state.maxJamFactor : undefined,
            ).then(({geoJson, isFiltered}) => {
                if (this.state.visibleLayer === mode) { // if removed by then
                    this.props.onTrafficLayerAdded(
                        geoJson,
                        mode!,
                        {fromDate: minDate, toDate: maxDate},
                        isFiltered);
                }
            }).catch((err: Error) => {
                this.props.showSnackbar(`Traffic Error: ${err.message}`, "error");
                console.error(err);
            }).then(() => {
                if (!this.state.visibleLayer || this.state.visibleLayer === mode) { // if removed by then
                    this.props.removeLoadingId(trafficLayerFetchId);
                    this.setState({isLoadingTraffic: false});
                }
            })
        );

        if (mode === "pm") {
            const pmDataFetchId = "pm-data-fetch";
            this.props.addLoadingId(pmDataFetchId);
            this.setState({isLoadingPM: true});
            promises.push(
                getPMGeoJsonWithAverages(
                    minDate,
                    maxDate,
                    mode === "pm" ? this.state.minPMValue : undefined,
                    mode === "pm" ? this.state.maxPMValue : undefined,
                ).then(({geoJson, isFiltered}) => {
                    if (this.state.visibleLayer) { // if removed by then
                        this.props.onPMLayerAdded(
                            geoJson,
                            mode!,
                            {fromDate: minDate, toDate: maxDate},
                            isFiltered
                        );
                    }
                }).catch((err: Error) => {
                    this.props.showSnackbar(`PM Error: ${err.message}`, "error");
                    console.error(err);
                }).then(() => {
                    this.props.removeLoadingId(pmDataFetchId);
                    this.setState({isLoadingPM: false});
                })
            );
        }

        return Promise.all(promises).then(() => {});
    }

    private isVisible(mode: TrafficVisibilityMode): boolean {
        return this.state.visibleLayer === mode;
    }

    private getSwitchChangedEventHandler(mode: TrafficVisibilityMode) {
        return (event: React.ChangeEvent<HTMLInputElement>) => {
            const checked = event.target.checked;
            this.setState({visibleLayer: checked ? mode : undefined});
            if (!mode || !checked) {
                this.props.onTrafficLayerRemoved();
            } else {
                this.handleSubmitClick(mode);
            }
        };
    }

    private shouldShowMinMaxValues() {
        return this.state.visibleLayer && ["speed", "jam"].includes(this.state.visibleLayer);
    }

    private shouldShowDayPicker() {
        return this.state.visibleLayer && ["co2", "pm"].includes(this.state.visibleLayer);
    }

    private getMinValue(): number | string {
        switch (this.state.visibleLayer) {
            case "speed":
                return numberOrEmptyString((this.state.minSpeed));
            case "jam":
                return numberOrEmptyString((this.state.minJamFactor));
        }

        return "";
    }

    private getMaxValue(): number | string {
        switch (this.state.visibleLayer) {
            case "speed":
                return numberOrEmptyString(this.state.maxSpeed);
            case "jam":
                return numberOrEmptyString(this.state.maxJamFactor);
        }

        return "";
    }

    private setMinValue(value?: string) {
        const val = stringToNumberOrUndefined(value);
        switch (this.state.visibleLayer) {
            case "speed":
                return this.setState({minSpeed: val});
            case "jam":
                return this.setState({minJamFactor: val});
        }
    }

    private setMaxValue(value?: string) {
        const val = stringToNumberOrUndefined(value);
        switch (this.state.visibleLayer) {
            case "speed":
                return this.setState({maxSpeed: val});
            case "jam":
                return this.setState({maxJamFactor: val});
        }
    }

    private handleColorByStrategyCheckbox(event: React.ChangeEvent<HTMLInputElement>) {
        const checked = event.target.checked;
        if (checked && !this.state.minDate && !this.state.maxDate) {
            const now = new Date();
            this.setState({
                maxDate: now,
                minDate: subWeeks(now, 1),
                colorSpeedByStrategy: checked,
            });
            setTimeout(() => {
                this.handleSubmitClick()
                    .then(() => this.props.onColorByStrategyChanged(checked));
            }, 50);
        } else {
            this.props.onColorByStrategyChanged(checked);
            this.setState({colorSpeedByStrategy: checked});
        }
    };

    private handleColorByAqiCheckbox(event: React.ChangeEvent<HTMLInputElement>) {
        const checked = event.target.checked;
        this.props.onColorByAqiChanged(checked);
        this.setState({colorPMByAqi: checked});
    };

    render() {
        return (<div>
            <FormGroup>
                <FormControlLabel
                    style={{padding: 10}}
                    control={<Switch
                        checked={this.isVisible("speed")}
                        onChange={this.getSwitchChangedEventHandler("speed")}
                        color={"primary"}
                    />}
                    label="Traffic Speed"
                />
                <Divider/>
                <FormControlLabel
                    style={{padding: 10}}
                    control={<Switch
                        checked={this.isVisible("jam")}
                        onChange={this.getSwitchChangedEventHandler("jam")}
                        color={"primary"}
                    />}
                    label="Traffic Jam"
                />
                <Divider/>
                <FormControlLabel
                    style={{padding: 10}}
                    control={<Switch
                        checked={this.isVisible("co2")}
                        onChange={this.getSwitchChangedEventHandler("co2")}
                        color={"primary"}
                    />}
                    label="CO2 Measurements"
                />
                <Divider/>
                <FormControlLabel
                    style={{padding: 10}}
                    control={<Switch
                        checked={this.isVisible("pm")}
                        onChange={this.getSwitchChangedEventHandler("pm")}
                        color={"primary"}
                    />}
                    label="PM Measurements"
                />
                <Divider/>
            </FormGroup>,
            <Grid container style={{padding: 0, display: this.state.visibleLayer ? "unset" : "none"}}>
                <Grid item xs={12} style={{margin: 10, display: this.state.visibleLayer === "speed" ? "unset" : "none"}}>
                    <FormControlLabel
                        style={{paddingTop: 10}}
                        control={<Checkbox
                            checked={this.state.colorSpeedByStrategy}
                            disabled={this.isLoading}
                            onChange={this.handleColorByStrategyCheckbox.bind(this)}
                            color={"primary"}
                        />}
                        label="Speed Classifications"
                    />
                </Grid>
                <Grid item xs={12} style={{margin: 10, display: this.state.visibleLayer === "pm" ? "unset" : "none"}}>
                    <FormControlLabel
                        style={{paddingTop: 10}}
                        control={<Checkbox
                            checked={this.state.colorPMByAqi}
                            disabled={this.state.isLoadingPM}
                            onChange={this.handleColorByAqiCheckbox.bind(this)}
                            color={"primary"}
                        />}
                        label="Air Quality Index"
                    />
                </Grid>
                <Grid item xs={12} style={{margin: 10}}>
                    <Typography variant={"h6"}>Filters</Typography>
                </Grid>
                <Grid item container xs={12} spacing={2} style={{padding: 8, margin: 0}}>
                    <Grid item xs={12} style={{display: this.shouldShowMinMaxValues() ? "unset" : "none"}}>
                        <TextField
                            type={"number"}
                            fullWidth
                            value={this.getMinValue()}
                            onChange={e => this.setMinValue(e.target.value)}
                            label={"Min"}
                        />
                    </Grid>
                    <Grid item xs={12} style={{display: this.shouldShowMinMaxValues() ? "unset" : "none"}}>
                        <TextField
                            type={"number"}
                            fullWidth
                            value={this.getMaxValue()}
                            onChange={e => this.setMaxValue(e.target.value)}
                            label={"Max"}
                        />
                    </Grid>
                    <MuiPickersUtilsProvider utils={DateFnsUtils}>
                        <Grid item xs={12} style={{display: !this.shouldShowDayPicker() ? "unset" : "none"}}>
                            <DateTimePicker
                                value={this.state.minDate}
                                onChange={(date) => this.setState({minDate: date})}
                                label="Min Date"
                                showTodayButton
                                todayLabel={"Now"}
                                clearable
                                autoOk
                                fullWidth
                            />
                        </Grid>
                        <Grid item xs={12} style={{display: !this.shouldShowDayPicker() ? "unset" : "none"}}>
                            <DateTimePicker
                                value={this.state.maxDate}
                                onChange={(date) => this.setState({maxDate: date})}
                                label="Max Date"
                                showTodayButton
                                todayLabel={"Now"}
                                clearable
                                autoOk
                                fullWidth
                            />
                        </Grid>
                        <Grid item xs={12} style={{display: this.shouldShowDayPicker() ? "unset" : "none"}}>
                            <DatePicker
                                value={this.state.day}
                                onChange={(date) => this.setState({day: date})}
                                label="Day"
                                showTodayButton
                                todayLabel={"Today"}
                                clearable
                                autoOk
                                fullWidth
                            />
                        </Grid>
                    </MuiPickersUtilsProvider>
                    <Grid item xs={12} style={{
                        textAlign: "center",
                        margin: 3
                    }}>
                        <Button
                            color={"primary"}
                            variant={"contained"}
                            disabled={this.isLoading}
                            onClick={() => this.handleSubmitClick()}
                        >
                            Submit
                        </Button>
                    </Grid>
                </Grid>
            </Grid>
        </div>);
    }
}
