import React from "react";
import {PopupDataInfo} from "./Popup";
import {PopupGrid, PopupGridItem} from "./PopupGrid";
import {
    trafficCo2ColorGradient,
    jamFactorColorGradient,
    pmTrafficColorGradient,
    speedColorGradient,
    ColorGradient
} from "../../common/ColorGradient";
import {
    BLACK_CSS,
    kgCo2PerDayFactorFromSpeed,
    formatTimestamp,
    getLabelAtValue,
    numberWithCommas,
    pmFactorFromSpeed,
    range,
    roundTo
} from "../../common/Helpers";
import {Typography} from "@material-ui/core";
import {addHours, endOfDay, startOfDay, subDays, subHours} from "date-fns";
import {getAveragesForFeatureOfInterest, getTrafficHistory} from "../../common/FrostApiProxy";
import {Bar, BarChart, CartesianGrid, Cell, Legend, Line, LineChart, ResponsiveContainer, Tooltip, XAxis, YAxis} from "recharts";
import {AxisDomain} from "recharts/types/util/types";
import {TrafficVisibilityMode} from "../Map";

type DateValue = {
    timestamp: number;
    value: number;
}

type Charts = {
    speedChart?: ChartData;
    jamChart?: ChartData;
    pmChart?: ChartData;
    co2Chart?: ChartData;
    selectedChart?: TrafficVisibilityMode;
}

type ChartData = {
    data: DateValue[];
    unit: string;
    axisTitle: string;
    yAxisDomain: AxisDomain;
    yAxisTickCount?: number;
    yAxisTicks?: (string | number)[];
}

type TrafficPopupState = {
    charts?: Charts;
    isLoading: boolean;
    fromDate: Date;
    toDate: Date;
}
type TrafficPopupProps = {
    info: PopupDataInfo;
}

export class TrafficPopupContent extends React.Component<TrafficPopupProps, TrafficPopupState> {

    constructor(props: TrafficPopupProps) {
        super(props);

        let {fromDate, toDate} = this.props.info.dateFilter;
        if (!fromDate && !toDate) {
            toDate = new Date();
            fromDate = subHours(toDate, 24);
        } else if (!fromDate && toDate) {
            fromDate = subHours(toDate, 24);
        } else if (fromDate && !toDate) {
            toDate = addHours(fromDate, 24);
        }

        this.state = {
            isLoading: false,
            fromDate: fromDate!,
            toDate: toDate!,
        };
    }

    componentDidMount() {
        if (!this.state.isLoading && !this.state.charts) {
            this.loadHistoryData();
        }
    }

    private loadHistoryData() {
        this.setState({isLoading: true});
        const {id} = this.props.info.trafficInfo!;
        const {fromDate, toDate} = this.state;
        const mode = this.props.info.trafficVisibilityMode!;
        if (mode === "speed" || mode === "jam") {
            getTrafficHistory(id, fromDate!, toDate!)
                .then(data => {
                    const speedChart: ChartData = {
                        axisTitle: "Speed",
                        unit: "km/h",
                        data: data.map(it => ({
                            timestamp: it.phenomenonTime.getTime(),
                            value: it.speed
                        })),
                        yAxisDomain: [0, (dataMax: number) => Math.max(dataMax, 100)],
                        yAxisTicks: [0, 20, 40, 60, 80, 100],
                    };
                    const jamChart: ChartData = {
                        axisTitle: "Jam Factor",
                        unit: "/10",
                        data: data.map(it => ({
                            timestamp: it.phenomenonTime.getTime(),
                            value: it.jam
                        })),
                        yAxisDomain: [0, 10],
                        yAxisTickCount: 6, // even numbers
                    };
                    this.setState({
                        charts: {
                            speedChart,
                            jamChart,
                            selectedChart: this.props.info.trafficVisibilityMode!
                        },
                        isLoading: false
                    });
                });
        } else {
            const last7Days = [6, 5, 4, 3, 2, 1, 0].map(d => ({
                startTime: subDays(startOfDay(fromDate), d),
                endTime: subDays(endOfDay(fromDate), d),
            }));
            const promises = last7Days.map(day =>
                getAveragesForFeatureOfInterest(
                    id,
                    1000,
                    day.startTime,
                    day.endTime,
                ).then(data => ({...data, ...day}))
            );
            Promise.all(promises).then(data => {
                const co2Chart: ChartData = {
                    axisTitle: "CO2 Emission",
                    unit: "kgCO2eq/day",
                    data: data.map(it => ({
                        timestamp: it.startTime.getTime(),
                        value: kgCo2PerDayFactorFromSpeed(
                            it.speed!,
                            this.props.info.trafficInfo?.segmentLength!
                        )
                    })),
                    yAxisDomain: [0, (dataMax: number) => Math.max(dataMax, 3000)],
                    yAxisTickCount: 7,
                };

                const pmChart: ChartData = {
                    axisTitle: "PM2.5",
                    unit: "kg/day",
                    data: data.map(it => ({
                        timestamp: it.startTime.getTime(),
                        value: pmFactorFromSpeed(it.speed!,
                            this.props.info.trafficInfo?.segmentLength!
                        )
                    })),
                    yAxisDomain: [0, (dataMax: number) => Math.max(dataMax, 300)],
                    yAxisTickCount: 6
                };

                this.setState({
                    charts: {
                        co2Chart,
                        pmChart,
                        selectedChart: this.props.info.trafficVisibilityMode!
                    },
                    isLoading: false
                });
            });
        }
    }

    private getGridItems(): PopupGridItem[] {
        const {trafficInfo, trafficVisibilityMode} = this.props.info;

        let items: PopupGridItem[] = [
            {key: "Id", value: trafficInfo?.id},
            {key: "Description", value: trafficInfo?.description},
        ];
        const averageStr = this.props.info.isFiltered ? " Average" : "";
        let textColor;
        switch (trafficVisibilityMode) {
            case "speed":
                textColor = speedColorGradient.getColorAtAsCssHex(trafficInfo?.speed, BLACK_CSS, true);
                items.push({key: "Speed" + averageStr, value: numberWithCommas(trafficInfo?.speed), textColor, unit: "km/h"});
                break;
            case "pm":
                textColor = pmTrafficColorGradient.getColorAtAsCssHex(trafficInfo?.pmFactor, BLACK_CSS, true);
                items.push({key: "PM2.5" + averageStr, value: numberWithCommas(trafficInfo?.pmFactor), textColor, unit: "kg/day"});
                break;
            case "co2":
                textColor = trafficCo2ColorGradient.getColorAtAsCssHex(trafficInfo?.co2Factor, BLACK_CSS, true);
                items.push({key: "CO2" + averageStr, value: numberWithCommas(trafficInfo?.co2Factor), textColor, unit: "kgCO2eq/day"});
                break;
            case "jam":
                textColor = jamFactorColorGradient.getColorAtAsCssHex(trafficInfo?.jamFactor, BLACK_CSS, true);
                items.push({key: "Jam" + averageStr, value: `${roundTo(trafficInfo?.jamFactor)}/10`, textColor});
                break;
        }
        if (trafficInfo?.quality) {
            items.push({key: "Result Quality", value: `${roundTo(trafficInfo.quality * 100)}%`});
        }
        return items;
    }

    private getChartTitle(): string {
        const {fromDate, toDate} = this.props.info.dateFilter;
        const parts = [];
        const dateTimeFormat = "dd.MM.yyyy HH:mm";
        const dayFormat = "dd.MM.yyyy";
        if (fromDate || toDate) {
            parts.push(`from ${formatTimestamp(this.state.fromDate.getTime(), dateTimeFormat)}`);
            parts.push(`to ${formatTimestamp(this.state.toDate.getTime(), dateTimeFormat)}`);
        }

        const period = parts.length ? parts.join(" ") : "from the last 24 hours";

        switch (this.props.info.trafficVisibilityMode) {
            case "speed":
                return `Speed data ${period}`;
            case "jam":
                return `Jam Factor data ${period}`;
            case "pm":
            case "co2":
                return `Daily emission rate a week from ${formatTimestamp(
                    this.state.fromDate.getTime(),
                    dayFormat
                )}`;
        }
        return `Data ${period}`;
    }

    private getSelectedChart(): ChartData | undefined {
        switch (this.state.charts?.selectedChart) {
            case "speed":
                return this.state.charts?.speedChart;
            case "jam":
                return this.state.charts?.jamChart;
            case "co2":
                return this.state.charts?.co2Chart;
            case "pm":
                return this.state.charts?.pmChart;
        }

        return undefined;
    }

    private getSelectedColorGradient(): ColorGradient | undefined {
        switch (this.state.charts?.selectedChart) {
            case "speed":
                return speedColorGradient;
            case "jam":
                return jamFactorColorGradient;
            case "co2":
                return trafficCo2ColorGradient;
            case "pm":
                return pmTrafficColorGradient;
        }

        return undefined;
    }

    private getMinMaxGradientValues() {
        const chartData = this.getSelectedChart();
        const minValue = chartData ? Math.min(...chartData.data.map(d => d.value)) : 0;
        const maxValue = chartData ? Math.max(...chartData.data.map(d => d.value)) : 0;
        const colorGradient = this.getSelectedColorGradient();

        return {minValue, maxValue, colorGradient};
    }

    render() {
        const selectedChart = this.getSelectedChart();
        const {minValue, maxValue, colorGradient} = this.getMinMaxGradientValues();
        return <div>
            <PopupGrid items={this.getGridItems()}/>
            <div style={{marginTop: 10}}>
                <Typography variant={"h6"}>{this.getChartTitle()}</Typography>
                {this.state.isLoading ? "Loading..." :
                    (() => {
                        const mode = this.props.info.trafficVisibilityMode;
                        if (mode === "speed" || mode === "jam")
                            return <ResponsiveContainer className={"area-chart-container"} height={350}>
                                <LineChart data={selectedChart?.data}>
                                    <defs>
                                        <linearGradient id="chartGradient" x1="0" y1="0" x2="0" y2="1">
                                            {range(0, 1, 10).map(v =>
                                                <stop key={`range-${v}`} offset={v}
                                                      stopColor={colorGradient?.getColorAtAsCssHex(
                                                          (1 - v) * (maxValue - minValue) + minValue, undefined, true
                                                      )}
                                                      stopOpacity={1}/>
                                            )}
                                        </linearGradient>
                                    </defs>
                                    <CartesianGrid strokeDasharray="3 3"/>
                                    <XAxis dataKey="timestamp" type={"number"}
                                           allowDataOverflow
                                           domain={["dataMin", "dataMax"]}
                                           tickFormatter={timestamp => formatTimestamp(timestamp)}/>
                                    <YAxis domain={selectedChart?.yAxisDomain}
                                           tickCount={selectedChart?.yAxisTickCount || 10}
                                           ticks={selectedChart?.yAxisTicks}
                                           tickFormatter={result => getLabelAtValue(roundTo(result))?.toString() || ""}/>
                                    <Legend/>
                                    <Tooltip
                                        labelFormatter={(timestamp) => formatTimestamp(timestamp, "dd.MM.yyyy HH:mm")}
                                        formatter={(result: number) => `${numberWithCommas(result)} ${selectedChart?.unit || ""}`}
                                    />
                                    <Line
                                        dataKey={"value"}
                                        dot={false}
                                        name={selectedChart?.axisTitle}
                                        strokeWidth={2}
                                        stroke={"url(#chartGradient)"}
                                        // @ts-ignore
                                        activeDot={({cx, cy, payload}) =>
                                            (<circle cx={cx}
                                                     cy={cy}
                                                     r={5}
                                                     stroke={"white"}
                                                     strokeWidth={1}
                                                     fill={colorGradient?.getColorAtAsCssHex(payload.value, "black", true)}
                                            />)
                                        }
                                    />
                                </LineChart>
                            </ResponsiveContainer>;

                        const data = selectedChart?.data;
                        const dayFormat = "E do. MMM";
                        return <ResponsiveContainer className={"area-chart-container"} height={350}>
                            <BarChart data={data}>
                                <defs>
                                    <linearGradient id="chartGradient" x1="0" y1="0" x2="0" y2="1">
                                        {range(0, 1, 10).map(v =>
                                            <stop key={`range-${v}`} offset={v}
                                                  stopColor={colorGradient?.getColorAtAsCssHex(
                                                      (1 - v) * (maxValue - minValue) + minValue, undefined, true
                                                  )}
                                                  stopOpacity={1}/>
                                        )}
                                    </linearGradient>
                                </defs>
                                <CartesianGrid strokeDasharray="3 3"/>
                                <XAxis dataKey="timestamp"
                                       tickFormatter={timestamp => formatTimestamp(timestamp, dayFormat)}/>
                                <YAxis
                                    domain={selectedChart?.yAxisDomain}
                                    tickCount={selectedChart?.yAxisTickCount || 10}
                                    ticks={selectedChart?.yAxisTicks}
                                    tickFormatter={result => getLabelAtValue(roundTo(result))?.toString() || ""}/>
                                <Legend/>
                                <Tooltip
                                    labelFormatter={(timestamp) => formatTimestamp(timestamp, dayFormat)}
                                    formatter={(result: number) => `${numberWithCommas(result)} ${selectedChart?.unit || ""}`}
                                />
                                <Bar
                                    dataKey={"value"}
                                    name={selectedChart?.axisTitle}
                                    fill={"url(#chartGradient)"}
                                >
                                    {data?.map((dateValue, index) =>
                                        <Cell key={`cell-${index}`}
                                              fill={colorGradient?.getColorAtAsCssHex(dateValue.value)}
                                        />
                                    )}
                                </Bar>
                            </BarChart>
                        </ResponsiveContainer>;
                    })()
                }
            </div>
        </div>;
    }
}
